/*
 * Small Wide World
 * by Joel R. Voss aka. Javantea
 * Copyright 2007 AltSci Concepts
 * October 25, 2008
 *
 * BF2 algorithm proof.
 * BF2 Na algorithm proof.
 * BF2 Nb algorithm proof.
 * BF2 Nc1 algorithm proof.
 */

/*
Description:
See bf2_secret1.txt.

based on brutemap1.cpp
by Javantea
June 28, 2007

TODO:

Rotation of full objects.
5c2
6c12345 (opposite, adjacent, same)
Nc1 is off by one I think...

*/

#include <cmath>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdlib.h>

using std::cout;
using std::cerr;
using std::endl;
using std::stringstream;

#include "sww.h"

#define PI 3.1415927f


/**
 * createMapNa(int nodes)
 * The incredible Na algorithm from SWW BF2.
 * The first node is a center node, the rest are mapped radially.
 * N refers to the number of nodes (nodes).
 * a refers to the number of connections between nodes not involving a (0).
 * Note: the second node (b) is never at the same angle, usually in the 
 * lower right.
 * 
 */
Map createMapNa(int nodes)
{
	int w = 640;
	int h = 480;
	Map a(NULL, "Test Na Map", w, h, 1);
	Vector2 c(w/2, h/2);
	Points &p = a.points;
	p.push_back(Point("a", 0+c.x, 0+c.y));
	int i;
	float theta = 2 * PI / (nodes - 1);
	float l = h / 4.0f;
	for(i = 1; i < nodes; i++)
	{
		char name[] = "a";
		name[0] += i;
		p.push_back(Point(name, l*sin(theta*i)+c.x, l*cos(theta*i)+c.y));
	}
	// a is connected to i
	for(i = 1; i < nodes; i++)
	{
		p[0].addConn(&(p[i]));
	}
	//a.center();
	return a;
}

/**
 * createMapNb(int nodes)
 * The Nb algorithm from SWW BF2.
 * N refers to the number of nodes (nodes).
 * b refers to the fact that there is one connection that does not involve a.
 * Because of this, every Nb should include an equilateral triangle (3b) and 
 * a number of tails with no deviation.
 * This will eventually become or be consumed by BF2 N_m algorithm since it's 
 * the same without (3b).
 * Note how it could be recursive.
 * 
 */
Map *createMapNb(int nodes, Map *init_map = NULL)
{
	int w = 640;
	int h = 480;
	Map *map = init_map;
	if(init_map == NULL) {
		//cout << "Initting a new map\n";
		// Initilize this dynamically.
		map = new Map(NULL, "Test Nb Map", w, h, 1);
	}
	Vector2 c(w/2, h/2);
	Points &p = map->points;
	float l = h / 4.0f;
	
	if(nodes < 3) {
		// 3b is the first Nb, so any N < 3 is an error case.
		return map;
	} else if(nodes >= 3) {
		/*
		h = sqrt(3/4)*l = 0.866025f * l
		b = l
		P[0] = 0,h/2.
		P[1] = b/2,-h/2.
		P[2] = -b/2,-h/2.
		*/
		float h = sqrt(0.75f) * l;
		float b = l;
		//Vector2 threeb[3] = { Vector2(0,h*0.5f), Vector2(b*0.5f,-h*0.5f), Vector2(-b*0.5f,-h*0.5f) };
		Vector2 threeb[3] = { Vector2(0,0), Vector2(b*0.5f,-h), Vector2(-b*0.5f,-h) };
		int i;
		for(i = 0; i < 3; i++)
		{
			char name[] = "a";
			name[0] += i;
			p.push_back(Point(name, threeb[i].x+c.x, threeb[i].y+c.y));
			//cout << "p1[" << i << "]\n";
		}
		/*// TODO: Why can't I do this?
		for(i = 0; i < 3; i++)
		{
			cout << "p[" << i << "] conn p[" << ((i+1) % 3) << "]\n";
			//p[i].addConn(&(p[((i+1) % 3)]));
		}*/
		//return map;
	} else {
		// nodes > 3, so add a triangle, then add the tails.
		//createMapNb(3, map);
	}
	//cout << "p has [" << p.size() << "]\n";
	
	int i;
	float phi = PI / 3.0f;
	// theta = (360-phi) / (N-2)
	float theta = (2 * PI - phi) / (nodes - 2);
	for(i = 3; i < nodes; i++)
	{
		char name[] = "a";
		name[0] += i;
		// theta_i = 30 + i * theta
		float theta_i = phi / 2.0f + (i-2) * theta;
		//cout << "theta_i[" << theta_i << " = " << phi << " / 2.0f + " << i << "* " << theta << "]\n";
		p.push_back(Point(name, l*sin(theta_i)+c.x, -l*cos(theta_i)+c.y));
		//cout << "p2[" << i << "]\n";
	}
	//cout << "p has [" << p.size() << "]\n";
	// Note ye who finds segfaults, if you don't have p[i], this will 
	// segfault very ugly. Darn, why C++, WHY?...
	// a is connected to i
	if(nodes != p.size()) {
		cout << "p is not be large enough: N (" << nodes << ") > p(" << p.size() << ").\n";
		return map;
	}
	for(i = 3; i < nodes; i++)
	{
		if(i >= p.size()) {
			cout << "p is not large enough: " << i << " vs " << p.size() << ".\n";
		}
		//cout << "p[" << 0 << "] conn p[" << i << "]\n";
		p[0].addConn(&(p[i]));
	}
	// TODO: Why do I have to do this here?
	for(i = 0; i < 3; i++)
	{
		//cout << "p[" << i << "] conn p[" << ((i+1) % 3) << "]\n";
		p[i].addConn(&(p[((i+1) % 3)]));
	}
	//a.center();
	return map;
}


/**
 * createMapNc(int nodes)
 * The Nc algorithm from SWW BF2.
 * N refers to the number of nodes (nodes).
 * c refers to the fact that there are two connections that does not involve a.
 * Because of this, every Nc should include a diamond (4c) and a number of 
 * tails with the only deviation being from which side.
 * This will eventually become or be consumed by BF2 N_m algorithm since it's 
 * the same without (4c).
 * 
 * 
 */
Map *createMapNc(int nodes, Map *init_map = NULL)
{
	int w = 640;
	int h = 480;
	Map *map = init_map;
	if(init_map == NULL) {
		//cout << "Initting a new map\n";
		// Initilize this dynamically.
		map = new Map(NULL, "Test Nc1 Map", w, h, 1);
	}
	Vector2 c(w/2, h/2);
	Points &p = map->points;
	float l = h / 4.0f;
	
	if(nodes < 4) {
		// 4c is the first Nc, so any N < 4 is an error case.
		return map;
	}
	/*
	h = sqrt(3/4)*l = 0.866025f * l
	b = l
	P[0] = 0,h/2.
	P[1] = b/2,-h/2.
	P[2] = -b/2,-h/2.
	*/
	float h1 = sqrt(0.75f) * l;
	float b = l;
	Vector2 fourc[4] = { Vector2(0,0), Vector2(b*0.5f,-h1), Vector2(-b*0.5f,-h1), Vector2(0,-2*h1) };
	int i;
	for(i = 0; i < 4; i++)
	{
		char name[] = "a";
		name[0] += i;
		p.push_back(Point(name, fourc[i].x+c.x, fourc[i].y+c.y));
		//cout << "p1[" << i << "]\n";
	}
	//cout << "p has [" << p.size() << "]\n";
	
	//int i;
	float phi = PI / 3.0f;
	// theta = (360-phi) / (N-2)
	float theta = (2 * PI - phi) / (nodes - 1);
	for(i = 4; i < nodes; i++)
	{
		char name[] = "a";
		name[0] += i;
		// theta_i = 30 + i * theta
		float theta_i = phi / 2.0f + (i-2) * theta;
		//cout << "theta_i[" << theta_i << " = " << phi << " / 2.0f + " << i << "* " << theta << "]\n";
		p.push_back(Point(name, l*sin(theta_i)+c.x, -l*cos(theta_i)+c.y));
		//cout << "p2[" << i << "]\n";
	}
	//cout << "p has [" << p.size() << "]\n";
	// Note ye who finds segfaults, if you don't have p[i], this will 
	// segfault very ugly. Darn, why C++, WHY?...
	// a is connected to i
	if(nodes != p.size()) {
		cout << "p is not be large enough: N (" << nodes << ") > p(" << p.size() << ").\n";
		return map;
	}
	for(i = 4; i < nodes; i++)
	{
		if(i >= p.size()) {
			cout << "p is not large enough: " << i << " vs " << p.size() << ".\n";
		}
		//cout << "p[" << 0 << "] conn p[" << i << "]\n";
		p[0].addConn(&(p[i]));
	}
	// TODO: Why do I have to do this here?
	for(i = 0; i < 3; i++)
	{
		//cout << "p[" << i << "] conn p[" << ((i+1) % 3) << "]\n";
		p[i].addConn(&(p[((i+1) % 3)]));
	}
	p[1].addConn(&(p[3]));
	p[3].addConn(&(p[2]));
	//a.center();
	return map;
}

void usage()
{
	cout << "Small Wide World BF2 algorithm \n";
	cout << "Usage: brutemap2_static [mapType] [maps]\n";
	cout << "mapType: which type of map to test, valid values: a, b.\n";
	cout << "maps: integer number of maps to render.\n";
}


int main(int argc, char **argv)
{
	char mapType = 'a';
	// "Wait, they don't love you like I love you" -- Yeah yeah yeahs
	int maps = 20;
	int i;
	
	for(i = 1; i < argc; i++)
	{
		if((strcmp("-h", argv[i]) == 0) || (strcmp("--help", argv[i]) == 0)) {
			usage();
			return 1;
		}
		if(i == 1) {
			mapType = argv[i][0];
		}
		if(i == 2) {
			maps = atoi(argv[i]);
		}
	}
	
	if(mapType == 'a') {
		for(i = 1; i < maps; i++)
		{
				Map m = createMapNa(i);
				string filename = "bf2_" + itostr(i) + "_a.svg";
				std::ofstream output(filename.c_str());
				output << m.str();
				output.close();
		}
	} else if(mapType == 'b') {
		for(i = 3; i < maps; i++)
		{
				Map *m = createMapNb(i);
				string filename = "bf2_" + itostr(i) + "_b.svg";
				std::ofstream output(filename.c_str());
				output << m->str();
				output.close();
				delete m;
		}
	} else if(mapType == 'c') {
		for(i = 4; i < maps; i++)
		{
				Map *m = createMapNc(i);
				string filename = "bf2_" + itostr(i) + "_c.svg";
				std::ofstream output(filename.c_str());
				output << m->str();
				output.close();
				delete m;
		}
	}
	
	return 0;
}
