#!/usr/bin/env python3
"""
objmap1.py
by Javantea

This module creates function descriptions from disassembly. The output can 
be mapped by graphical programs (mapmake1).

For open source programs, it is possible to do this by parsing the source 
code, but it is a much larger task. This is a very easy task.

"""
from __future__ import print_function
import re
from random import seed
import mapmake1
try:
	import sww_optimize1
except ImportError:
	print("Warning: SWW Optimization has missing dependencies. Try running sww_optimize1.py to find out what.", file=sys.stderr)
	sww_optimize1 = None
#end try

def mapDisasm(filename):
	"""
	mapDisasm takes a dump from objdump -d and turns it into a useful list of 
	functions and the calls that they make.
	"""
	a = open(filename, 'r').read()
	b = a.split('\n')
	named_functions = []
	called_funcs = set()
	for line in b:
		func_re = re.match(r'^([0-9a-f]+) \<([^\>]+)\>:', line)
		if func_re != None:
			named_functions.append(list(func_re.groups()))
			continue
		#end if
		call_re = re.match(r'^ +([0-9a-f]+):\t.{21}\tcallq?   ?([0-9a-f]+) \<([^\>]+)\>', line)
		if call_re != None:
			addr, call_addr, call_name = call_re.groups()
			called_funcs.add(int(call_addr, 16))
		#end if
	#next line
	
	named_func_addresses = [int(addr, 16) for addr, name in named_functions]
	cfsort = list(set(list(called_funcs) + named_func_addresses))
	cfsort.sort()
	functions = []
	for cfaddr in cfsort:
		if cfaddr in named_func_addresses:
			name = named_functions[named_func_addresses.index(cfaddr)][1]
		else:
			name = 'sub_%x' % cfaddr
		#end if
		functions.append(['%x' % cfaddr, name])
	#next cfaddr

	if len(functions) <= 0:
		return []
	#end if
	
	func_i = 0
	func = functions[0]
	if len(functions) > 1:
		next_func = functions[1]
	else:
		next_func = ['0', 'unknown']
	#end if
	for line in b:
		call_re = re.match(r'^ +([0-9a-f]+):\t.{21}\tcallq? *([0-9a-f]+) \<([^\>]+)\>', line)
		if call_re != None:
			ff = call_re.groups()
			addr, call_addr, call_name = ff
			addri = int(addr, 16)
			while addri >= int(next_func[0], 16):
				func_i += 1
				func = next_func
				if func_i+1 >= len(functions):
					break
				#end if
				next_func = functions[func_i+1]
			#loop
			func.append(ff)
			continue
		#end if
	#next line
	
	return functions
#end def mapDisasm(filename)

def objmap(filename, seed_id = 2345, callersOnly = False):
	seed(seed_id)
	m = mapmake1.Map()
	#'/home/jvoss/recent/pfm/projects/objmap/python1.dis'
	f = mapDisasm(filename)
	
	"""# the format of functions is:
	# [address, name, (call addr, call pos, call name), ...]
	for func in f:
		print func[1] + ': (',
		for call in func[2:]:
			print call[2] + ',',
		#next call
		print ')'
	#next func
	return"""
	
	# The variable callersOnly filters out those functions that make no calls.
	# The graph removes dots which are less interesting and far more common in 
	# crazy binaries. The map will still show dots where the caller's called 
	# item has been removed. TODO: write code to filter only caller/called.
		
	i = 0
	p = []
	names = {}
	max_func = ''
	max_calls = 0
	for func in f:
		if (not callersOnly) or len(func) > 2:
			p.append(mapmake1.Point(func[1]))
			names[func[1]] = i
			i += 1
		#end if
		if len(func) - 2 > max_calls:
			max_calls = len(func) - 2
			max_func = func[1]
		#end if
	#next func
	i = 0
	for func in f:
		for call in func[2:]:
			j = call[2].split('-')
			if(len(j) == 1): j = call[2].split('+')
			if j[0] in names:
				conn = names[j[0]]
				p[i].addConn(p[conn])
			else:
				if not callersOnly: print('Missing name: call', j[0])
			#end if
		#next call
		if (not callersOnly) or len(func) > 2:
			i += 1
		#end if
	#next func
	
	m.addPoints(p)
	for i in range(4):
		m.randomize()
		m.sort1()
		#a.sort1()
		#a.sort1()
		#a.sort1()
		#a.sort1()
	#next i
	if sww_optimize1:
		print("Nodes:", len(p))
		# This is really useful to make the output viewable, but it's slow.
		# Split the map into separate maps.
		maps = m.split()
		print("Maps:", len(maps))
		used_space = [0, 40]
		max_y_this_row = 0
		for m2set in maps:
			m2 = mapmake1.Map(list(m2set))
			#print(', '.join([p.name for p in m2.points]))
			# Don't need to deal with the single items.
			if len(m2set) > 1:
				lj_conf = sww_optimize1.LJ_Potential(120)
				sww_optimize1.OptimizeBasinHopping(m2, lj_conf, 100, 30, 120)
			#end if
			# Now we need to pack these maps into a single map.
			# There are better parcel packing algorithms, but this one is mine.
			rect = m2.bounds()
			# Move it to (5, 5)
			y_use = rect[3] - rect[1] + 20 # grab an extra 20 pixels.
			if y_use > max_y_this_row:
				max_y_this_row = y_use
			#end if
			if used_space[0]+(rect[2] - rect[0]) > m.width:
				if used_space[0] != 0:
					# Use the next line instead of this one.
					used_space[1] += max_y_this_row
					used_space[0] = 0
				#end if
			#end if
			#print('stack:', 5+used_space[0], 5+used_space[1])
			#print('size: %3.1f %3.1f' % (rect[2] - rect[0], rect[3] - rect[1]))
			m2.move(-rect[0]+5+used_space[0], -rect[1]+5+used_space[1])
			# columns, then rows.
			used_space[0] += rect[2] - rect[0] + 150
			if used_space[0] > m.width:
				# Use the next row.
				used_space[1] += max_y_this_row
				used_space[0] = 0
			#end if
		#next m2
		#m.center()
	#end if
	m.max_func = max_func
	m.max_calls = max_calls
	
	return m
#end def objmap(filename, [seed_id])

def main(seed_id = 2345):
	filename = 'data/viawget.dis'
	callersOnly = False
	
	from sys import argv
	if '-c' in argv:
		argv.pop(argv.index('-c'))
		callersOnly = True
		print(argv)
	#end if
	if len(argv) > 1:
		filename = argv[1]
	#end if
	
	m = objmap(filename, seed_id, callersOnly)
	print(m)
	print("<!-- max_func =", m.max_func, ", max_calls =", m.max_calls, " -->")
#end def main(seed_id)

if __name__ == '__main__':
	main()
#end if



