var tau = 2*Math.PI;

/*
Map(points, width, height, line_width, line_color)
Base class for all maps (friend, object, route).
Write the hard math and physics here once and use it by calling once.
*/
function Map(points, width, height, line_width, line_color) {
	if(points == undefined) { points = []; }
	if(width == undefined) { width = 640; }
	if(height == undefined) { height = 480; }
	if(line_width == undefined) { line_width = 2; }
	if(line_color == undefined) { line_color = '#000000'; }
	this.points = points;
	this.width  = width;
	this.height = height;
	this.line_width = line_width;
	this.line_color = line_color;
	this.selectedPoint = null;
	this.title = 'Small Wide World Map';
	this.desc = 'Test of drawing a network map with svg';
}

// Make it a class.
new Map();

Map.prototype.addPoint = function(point) {
	this.points.push(point);
}

Map.prototype.addPoints = function(points) {
	this.points.extend(points);
}

/*if (debug == undefined)
{
	// Silent but deadly.
	debug = function() {}
}*/

/*
 * This randint doesn't care if you get the params backwards.
 * Therefore it functions same in all sane calls, functions better in insane 
 * calls.
 * Similar to python random.randint(x, y)
 */
function randint(x, y) {
	var dx = Math.abs(y - x) + 1;
	var minXY = Math.min(x, y);
	return Number((minXY + (Math.random() * (dx-1))).toFixed());
}

Map.prototype.randomize = function(amount) {
	// Deal with optional parameters.
	if(amount == undefined) { amount = 180; }
	var i;
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		point.x += randint(-amount, amount)
		point.y += randint(-amount, amount)
	} //#next point
} //#end Map.prototype.randomize

Map.prototype.drag = function(x, y) {
	if (this.selectedPoint == null) {
		return false;
	}
	var r = new Vector2(x, y);
	var d_len = (r.sub(this.selectedPoint)).length();
	if (d_len > 10) {
		this.selectedPoint.x = x;
		this.selectedPoint.y = y;
		return true;
	}
	return false;
} //#end Map.prototype.drag

Map.prototype.click = function(x, y) {
	var cp = new Vector2(x, y);
	var min_dr = -1.0;
	var min_dr_p = null;
	var i;
	for (i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		dr = cp.sub(point);
		dr_len = dr.length();
		if ((min_dr < 0) || (dr_len < min_dr)) {
			min_dr = dr_len;
			min_dr_p = point;
		}
	} //#next point
	if (min_dr_p != null) {
		//print("Setting color!")
		min_dr_p.setColor('#0000cc');
		if (this.selectedPoint != null) {
			this.selectedPoint.setColor('#ccc');
		}
		this.selectedPoint = min_dr_p;
	}
} //#end Map.prototype.click

/*
 * Map.getNamed
 * Only available in JavaScript for now.
 * Aides Conn('a', 'b')
 */
Map.prototype.getNamed = function(s) {
	var i;
	for(i = 0; i < this.points.length; i++)
	{
		if(this.points[i].name == s)
			return this.points[i];
	}
	return null;
}

/*
 * Map.sort1()
 * Bring all points together.
 */
Map.prototype.sort1 = function() {
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			var dr = conn.sub(point);
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			point.x += dr.x * 0.4;
			point.y += dr.y * 0.4;
		} //#next conn
	} //#next point
} //#end Map.prototype.sort1

/*
 * Map.avgLength()
 * Expensive function tot find average length.
 * O(n**2)
 */
Map.prototype.avgLength = function() {
	var avgLen = 0;
	var conns = 0;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			var dr = conn.sub(point);
			//#dr_len = sqrt(dx*dx + dy*dy)
			var dr_len = dr.length();
			avgLen += dr_len;
			conns += 1;
		}
	}
	//debug("Average Length: " + (avgLen / conns));
	if (conns == 0)
		return 80;
	return avgLen / conns;
}

/*
 * Map.sort2(dist1, f)
 * Use springs to decide the position of each point.
 * Moves the whole thing to the lower right due to a bug.
 */
Map.prototype.sort2 = function(dist1, f) {
	// Deal with optional parameters.
	if(dist1 == undefined) { dist1 = this.avgLength(); }
	if(f == undefined) { f = 0.5; }
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			//#(dx, dy) = (conn.x - point.x, conn.y - point.y)
			var dr = conn.sub(point);
			//#dr_len = sqrt(dx*dx + dy*dy)
			var dr_len = dr.length();
			//#print '	dr:', dr
			if (dr_len == 0.0) {
				point.x += 5.0;
				point.y += 5.0;
				dr_len = Math.sqrt(5*5 + 5*5);
				//#print 'zero dr is now:', dr
			}
			//#print 'non-zero dr:', dr
			var m = 1;
			var a_len = (f/m) * (dr_len - dist1);
			//#dr_hat = (dx/dr, dy/dr)
			var dr_hat = dr.div(dr_len);
			//#a = (a_len*dr_hat[0], a_len*dr_hat[1]/m)
			var a = dr_hat.mul(a_len / m);
			dt = 0.6;
			//#dp = (0.5*dt*dt*a[0], 0.5*dt*dt*a[1])
			var dp = a.mul(0.5 * dt * dt);
			point.x += dp.x;
			point.y += dp.y;
			//debug(i + " " + a_len + " " + dr_hat.toString());
		} //#next conn
	} //#next point
}//#end Map.prototype.sort2

/*
 * Map.sort3
 * Pull all nodes to the center of their connections.
 */
Map.prototype.sort3 = function() {
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		if (point.conns.length <= 1) continue;
		//#avgx = 0
		//#avgy = 0
		var avg = new Vector2(0, 0);
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			//#avgx += conn.x
			//#avgy += conn.y
			avg = avg.add(conn);
		} //#next conn
		//#point.x = avgx / point.conns.length
		//#point.y = avgy / point.conns.length
		var p = avg.div(point.conns.length);
		point.x = p.x;
		point.y = p.y;
	} //#next point
}//#end Map.prototype.sort3

function checkPointInConns(point, conns) {
	var i;
	for(i = 0; i < conns.length; i++) {
		if (conns[i].b == point)
			return true;
	}
	return false;
}

/*
 * Map.sort4()
 * Use a style of reverse attack on non-conns that are close.
 * FIXME: broken
 */
Map.prototype.sort4 = function() {
	var i;
	var j;
	var dr, dr_len;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		for (j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (checkPointInConns(nonconn, point.conns)) {
				continue;
			}
			dr = point.sub(nonconn);
			dr_len = dr.length();
			//Console.log("dr_len = " + dr_len);
			//#print 'dr_len =', dr_len
			if (dr_len < 50) {
				if (dr_len == 0) {
					dr = new Vector2(1, 1);
					dr_len = Math.sqrt(2);
				}
				dr = dr.mul(4.0 / dr_len);
				point.x += dr.x;
				point.y += dr.y;
			}
		} //#next nonconn
	} //#next point
}//#end Map.prototype.sort4

/*
 * Map.sort5()
 * This is obviously the best algorithm ever. Hooray for the power 
 * of awesomeness.
 * FIXME: broken
 */
Map.prototype.sort5 = function() {
	var i;
	var j;
	var dr, dr_len;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var enemies_pos = new Vector2(0, 0);
		var enemies_len = 0;
		var friends_pos = new Vector2(0, 0);
		for (j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (checkPointInConns(nonconn, point.conns)) {
				friends_pos = friends_pos.add(nonconn);
				continue;
			}
			dr = point.sub(nonconn);
			dr_len = dr.length();
			//#print 'dr_len =', dr_len
			if (dr_len < 50) {
				enemies_pos = enemies_pos.add(nonconn);
				enemies_len += 1;
			}
		} //#next nonconn
		enemies_pos = enemies_pos.div(enemies_len);
		friends_pos = friends_pos.div(point.conns.length);
		dr = friends_pos.mul(2).sub(enemies_pos);
		//# Our position looks like this: enemies ---- friends ---- me
		point.x = dr.x;
		point.y = dr.y;
		//debug(point.toString());
		//debug(' &nbsp; &nbsp;' + '(' + (enemies_len) + ')', enemies_pos);
		//debug(' &nbsp; &nbsp;' + '(' + (point.conns.length) + ')', friends_pos);
	} //#next point
}//#end Map.prototype.sort5

/*
 * Map.sort6()
 * A new algorithm based on sort5 but with added validity information.
 * FIXME: broken
 */
Map.prototype.sort6 = function() {
	var i;
	var j;
	var dr;
	var validity = new Array(this.points.length);
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var enemies_pos = new Vector2(0, 0);
		var enemies_len = 0;
		var friends_pos = new Vector2(0, 0);
		var friends_len = 0;
		for(j = 0; j < this.points.length; j++) {
			var nonconn = this.points[j];
			if (!validity[j]) {
				continue;
			}
			if (checkPointInConns(nonconn, point.conns)) {
				friends_pos = friends_pos.add(nonconn);
				friends_len += 1;
				continue;
			}
			/*
			dr = point.sub(nonconn)
			dr_len = dr.length()
			print 'dr_len =', dr_len
			if (dr_len < 150) {
			*/
			enemies_pos = enemies_pos.add(nonconn);
			enemies_len += 1;
			//}
		} //#next nonconn
		if (enemies_len > 1) enemies_pos = enemies_pos.div(enemies_len);
		if (friends_len > 1) friends_pos = friends_pos.div(friends_len);
		if ((enemies_len > 0) || (friends_len > 0)) {
			dr = friends_pos.mul(1.7).sub(enemies_pos);
			if ((enemies_len == 0) || (dr.sub(friends_pos).length() < 1) || (dr.sub(enemies_pos).length() < 1)) {
				//# Fix anything by using a random angle.
				var theta = randint(0, 360) * (Math.PI / 180.0);
				dr = new Vector2(dr.x + 30 * Math.cos(theta), dr.y + 30 * Math.sin(theta));
			}
		} else {
			dr = new Vector2(0, 0);
		}
		//# Our position looks like this: enemies ---- friends ---- me
		point.x = dr.x + (randint(0, 100) / 4.0);
		point.y = dr.y + (randint(0, 100) / 4.0);
		//#print point
		//#print '\t' + '(' + str(enemies_len) + ')', enemies_pos
		//#print '\t' + '(' + str(friends_len) + ')', friends_pos
		validity[i] = true;
	} //#next point
} //#end Map.prototype.sort6

/*
 * Map.center()
 * Centers the graph on the screen.
 */
Map.prototype.center = function() {
	if (this.points.length == 0) {
		return;
	}
	var mean = new Vector2(0, 0);
	var i;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		mean = mean.add(point);
	} //#next point
	mean = mean.div(1.0 * this.points.length);
	var goto = new Vector2(this.width/2.0, this.height/2.0);
	var diff = goto.sub(mean);
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		point.x += diff.x;
		point.y += diff.y;
	} //#next point
}; //#end Map.prototype.center

/*
 * Map.sortBranch(r0=40)
 * A brilliant sorting method that works well with branches but not loops.
 * r0 is the length of the bonds set by the algorithm.
 * roundTry is necessary only for loops.
 */
Map.prototype.sortBranch = function(r0, debugSort) {
	if (r0 == undefined) { r0 = 40; }
	if (debugSort == undefined) { debugSort = false; }
	var nodes = this.points.length;
	//debug("sortBranch(" + nodes + ")");
	if (nodes < 1) {
		//# Congrats, you're guaranteed to be sorted.
		return;
	}
	//# This is our algorithm... Written on the whiteboard last night. Understood many years ago but unable to write the code. Until now!
	var knownPos = Array(nodes).fill(false); //[false] * nodes;
	knownPos[0] = true;
	this.points[0].x = this.width / 2;
	this.points[0].y = this.height / 2;
	//# v1 in spherical coordinates.
	//#v1 = (r1, theta1) == (r1*cos(theta1), r1*sin(theta1))
	var v1 = [r0, 0];
	//# -v1 in spherical coordinates
	//var neg_v1 = [r0, v1[1] + Math.PI];
	
	for (roundTry = 0; roundTry < nodes; roundTry++) {
		for(i = 0; i < nodes; i++) {
			//#print(self.points[i].name, self.points[i].conns)
			givens = [];
			unknowns = [];
			//# Set the position of its connections.
			// for conn in this.points[i].conns:
			for (k = 0; k < this.points[i].conns.length; k++) {
				var conn = this.points[i].conns[k].b;
				var j = this.points.indexOf(conn);
				if(knownPos[j]) {
					givens.push(j);
					continue;
				} //#end if
				unknowns.push(j);
			} //#next conn
			if (knownPos[i]) {
				//# First solve the easy cases, then work your way up.
				if(unknowns.length == 0) {
					//# Everything is known, we're good.
					if (debugSort) debug('Yay ' + this.points[i].x + " " + this.points[i].y)
					continue;
				} //#end if
				if(givens.length == 0) {
					//# All unknowns, so use the simplex method.
					var n_conns = unknowns.length;
					var theta0 = tau / n_conns; //# tau = 2*pi
					var x = 0;
					// for j in unknowns:
					for(k = 0; k < unknowns.length; k++) {
						var j = unknowns[k];
						//# Spherical coordinates
						var theta_j = v1[1] + (theta0 * (x+1));
						//#v_j = (r0, theta_j)
						this.points[j].x = this.points[i].x + (r0*Math.cos(theta_j)) 
						this.points[j].y = this.points[i].y + (r0*Math.sin(theta_j))
						knownPos[j] = true;
						if (debugSort) debug('Simple ' + x + " " + j + " " + this.points[j].x + " " + this.points[j].y + " " + theta_j)
						x += 1;
					} //#next j
					continue;
				} //#end if
				if(givens.length == 1) {
					//# Very easy, create a v1 and do the same as above
					var v1i_cartesian = this.points[i].sub(this.points[givens[0]]);
					var v1i_theta = Math.atan2(v1i_cartesian.y, v1i_cartesian.x);
					//#v1i_spherical = (v1i_cartesian.length(), v1i_theta)
					//#neg_v1i_spherical = (v1i_cartesian.length(), v1i_theta+pi)
					var n_conns = unknowns.length + 1;
					var theta0 = tau / n_conns; //# tau = 2*pi
					var x = 0;
					// for j in unknowns:
					for(k = 0; k < unknowns.length; k++) {
						var j = unknowns[k];
						//# Spherical coordinates
						if(n_conns == 2) {
							theta_j = v1i_theta;
						} else {
							theta_j = v1i_theta+Math.PI + (theta0 * (x+1));
						} //#end if
						//#v_j = (r0, theta_j)
						this.points[j].x = this.points[i].x + (r0*Math.cos(theta_j))
						this.points[j].y = this.points[i].y + (r0*Math.sin(theta_j))
						knownPos[j] = true;
						if (debugSort) debug('Simple1 ' + x + " " + j + " " + this.points[j].x + " " + this.points[j].y + " " + theta_j + " " + 'v1it' + " " + v1i_theta)
						x += 1;
					} //#next j
				} //#end if
				if (debugSort) debug("Not sure what to do for: given: " + givens.length + ", unknown: " + unknowns.length + "")
				//# Take the list of givens and create a perfect placement for our node
			} else {
				//debug("Not known pos " + i + " " + knownPos[i])
				if(givens.length == 0) {
					//# We don't know any of our neighbors yet, nope.
					continue;
				} //#end if
				if(givens.length == 1) {
					//# Use v1 to choose our position
					knownPos[i] = true;
					var j = givens[0];
					if (debugSort) debug(i + ' using v_' + j)
					var theta_j = v1[1];
					this.points[i].x = this.points[j].x + (r0*Math.cos(theta_j));
					this.points[i].y = this.points[j].y + (r0*Math.sin(theta_j));
					continue;
				} //#end if
				if (debugSort) debug('avg')
				//# Average of its peers
				var totalPos = [0, 0];
				// for j in givens:
				for(k = 0; k < givens.length; k++) {
					var j = givens[k];
					totalPos[0] += this.points[j].x;
					totalPos[1] += this.points[j].y;
				} //#next j
				var avg = [totalPos[0] / givens.length, totalPos[1] / givens.length];
				this.points[i].x = avg[0];
				this.points[i].y = avg[1];
			} //#end if
			// if not false in knownPos:
			if (knownPos.indexOf(false) == -1) break;
		} //#next i
		// if not false in knownPos:
		if (knownPos.indexOf(false) == -1) {
			if (debugSort) debug("Took " + (roundTry + 1) + " tries: " + knownPos.toString())
			break;
		} //#end if
	} //#next roundTry
}; //#end def sortBranch(r0=40)
	
/*
 * Map.squash()
 * A very strange algorithm, not sure why it was written.
 * 
 */
Map.prototype.squash = function() {
	var mean = new Vector2(0, 0);
	for(i = 0; i < this.points.length; i++) {
		//#mean[0] += point.x
		//#mean[1] += point.y
		mean = mean.add(point);
	} //#next point
	//#mean[0] /= (1.0 * this.points.length)
	//#mean[1] /= (1.0 * this.points.length)
	mean = mean.div(1.0 * this.points.length);
	//#goto = [this.width/2.0, this.height/2.0]
	var goto = new Vector2(this.width/2.0, this.height/2.0);
	//#diff = [goto[0]-mean[0], goto[1]-mean[1]]
	var diff = goto.sub(mean);
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		//#point.x += diff[0]
		//#point.y += diff[1]
		for (j = 0; j < point.conns.length; j++) {
			// FIXME: This seems totally wrong.
			var conn = point.conns[i];
			point.x += diff.x;
			point.y += diff.y;
		} //#next conn
	} //#next point
}; //#end Map.prototype.squash

/*
 * Map.toString()
 * Returns an SVG of the Map.
 */
Map.prototype.toString = function() {
	return this.drawSvg();
}; //#end Map.prototype.__str__

Map.prototype.drawSvg = function() {
	var output = '<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">';
	output += '\n<svg width="' + str(this.width) + '" height="' + str(this.height) + '" viewBox="0 0 ' + str(this.width) + ' ' + str(this.height) + '" xmlns="http://www.w3.org/2000/svg" version="1.1" xmlns:xlink="http://www.w3.org/1999/xlink">\n';
	output += '<title>' + this.title + '</title>\n';
	output += '<desc>' + this.desc + '</desc>\n';
	output += '<rect x="1" y="1" width="' + str(this.width-2) + '" height="' + str(this.height-2) + '" fill="none" stroke="blue" />\n';
	
	var i;
	var j;
	for(i = 0; i < this.points.length; i++) {
		var point = this.points[i];
		var tpos = point.add(new Vector2(8, 8));
		//# Draw each point
		//#output += "<!--<text x='" + ('%3.3f' % tpos.x) + "' y='" + ('%3.3f' % tpos.y) + "'>" + point.name + "</text>-->\n"
		output += "<circle cx='" + ('%3.3f' % point.x) + "' cy='" + ('%3.3f' % point.y) + "' r='" + ('%3.1f' % point.size) + "' fill='" + point.color + "' stroke='grey' stroke-width='1' />\n";
		//# Draw each conn
		for (j = 0; j < point.conns.length; j++) {
			var conn = point.conns[j].b;
			//#output += '<!-- ' + str(point) + ' to ' + str(conn) + '-->\n'
			output += "<path d='M " + ('%3.3f' % point.x) + " " + ('%3.3f' % point.y) + " L " + ('%3.3f' % conn.x) + " " + ('%3.3f' % conn.y) + "' fill='none' stroke='" + str(this.line_color) + "' stroke-width='" + str(this.line_width) + "' />\n";
		} //#next conn
	} //#next point
	//# A layer of text above the connections.
	output += '<g>';
	for(i = 0; i < this.points.length; i++) {
		var tpos = point.add(new Vector2(8, 8));
		if (point.url != '') {
			output += "<a xlink:href='" + point.url.replace('&', '&amp;') + "'>";
		}
		output += "<text x='" + ('%3.3f' % tpos.x) + "' y='" + ('%3.3f' % tpos.y) + "' font-size='11'>" + point.name.replace('&', '&amp;') + "</text>\n";
		if (point.url != '') {
			output += "</a>\n";
		}
	}//#next point
	output += '</g>';
	output += '</svg>';
	return output;
}; //#end Map.prototype.drawSvg
//#end class Map

/*
 * Vector2(x, y)
 * 2D Vector implementation in JavaScript.
 * JavaScript lacks operator override, so we have methods instead.
 */
function Vector2(x, y)
{
	if(x == undefined) { x = 380.0; }
	if(y == undefined) { x = 180.0; }
	this.x = new Number(x);
	this.y = new Number(y);
}
new Vector2();

Vector2.prototype.toString = function() {
	return (this.x.toFixed(3)) + ', ' + (this.y.toFixed(3))
}; //#end Vector2.prototype.toString

Vector2.prototype.add = function(vec) {
	return new Vector2(this.x + vec.x, this.y + vec.y);
};
Vector2.prototype.sub = function(vec) {
	return new Vector2(this.x - vec.x, this.y - vec.y)
};
Vector2.prototype.mul = function(flt) {
	return new Vector2(this.x * flt, this.y * flt);
};

Vector2.prototype.div = function(flt) {
	//#if (flt == 0) {
	//#	return Vector2(0, 0)
	//debug("(" + this.x + ", " + this.y + ") / " + flt);
	return new Vector2(this.x / flt, this.y / flt);
};

//#Vector2.prototype.__iadd__ = function(vec) {:
//#	v = new Vector2(this.x + vec.x, this.y + vec.y)
//#	return v

//#Vector2.prototype.__idiv__ = function(flt) {:
//#	v = new Vector2(this.x / flt, this.y / flt)
//#	return v

Vector2.prototype.length = function() {
	return Math.sqrt(this.x * this.x + this.y * this.y);
};
//#end class Vector2

// Point2 Subclasses Vector2
function Point2(name, conns, x, y, color, url, size)
{
	// Deal with a bunch of optional params.
	if(x == undefined) { x = 380.0; }
	if(y == undefined) { y = 180.0; }
	if(color == undefined) { color = '#ccc'; }
	if(url == undefined) { url = ''; }
	if(size == undefined) { size = 4; }
	Vector2.call(this, x, y);
	//Vector2.__init__(self, x, y)
	this.name = name;
	this.url = url;
	this.conns = conns;
	this.color = color;
	this.size = size;
	this.visual = null;
	this.text = null;
	if (conns == null) this.conns = [];
}
Point2.prototype = Object.create(Vector2.prototype);
new Point2();

Point2.prototype.toString = function() {
	return this.name.toString() + ' ' + (this.x.toFixed(3)) + ', ' + (this.y.toFixed(3));
}; //#end Point2.prototype.__str__ = function() {

/*
 * Point2.addConn(conn, reConn, connObj)
 * Different from mapmake1.py due to the Conn object.
 * reConn is optional like in Python.
 * connObj is optional and should be left empty (undefined) so that the 
 * reconnection works.
 */
Point2.prototype.addConn = function(conn, reConn, connObj) {
	if(reConn == undefined) { reConn = true; }
	// Already connected, don't add duplicates.
	if(checkPointInConns(conn, this.conns)) { return; }
	var ourConnObj = new Conn(this, conn);
	if(connObj != undefined){
		// Connect the visuals so we don't draw twice.
		//ourConnObj.visualc = connObj.visualc;
	}
	//#print 'adding conn?', this.name, conn
	this.conns.push(ourConnObj);
	// Give them the same object, so when it renders, it doesn't render two 
	// objects, it renders the same object twice if you get what I mean.
	if(reConn) conn.addConn(this, false, ourConnObj);
};
Point2.prototype.setColor = function(value) {
	/* Set color to a hex value like #ff0000 */
	this.color = value;
};

/*
 * Point2::draw() -- update the screen.
 * Only available in JavaScript for obvious reasons.
 */
Point2.prototype.draw = function(display)
{
	// Draw it.
	if(this.visual === null)
	{
		var svg = 'http://www.w3.org/2000/svg';
		this.visual = document.createElementNS(svg, 'circle');
		this.visual.setAttributeNS(null, 'stroke', '#999');
		this.visual.setAttributeNS(null, 'stroke-width', '2');
		this.visual.setAttributeNS(null, 'stroke-opacity', '1');
		this.visual.setAttributeNS(null, 'fill', this.color);
		this.visual.onmousedown = function() { return false; }
		this.text = document.createElementNS(svg, 'text');
		this.text.setAttributeNS(null, 'font-family', 'Droid Sans Mono;monospace');
		this.text.setAttributeNS(null, 'font-size', '12');
		this.text.setAttributeNS(null, 'style', "fill:#000");
		// TODO: make this editable.
		//this.text.contentEditable = true;
		this.text.onmousedown = function() { return false; }
		this.text.innerHTML = htmlentities(this.name.toString());
		display.appendChild(this.visual);
		display.appendChild(this.text);
	}

	this.visual.setAttributeNS(null,'cx', this.x);
	this.visual.setAttributeNS(null,'cy', this.y);
	this.visual.setAttributeNS(null,'r', 10);
	this.text.setAttributeNS(null,'x', this.x+14);
	this.text.setAttributeNS(null,'y', this.y+14);

	var i;
	for(i = 0; i < this.conns.length; i++)
	{
		this.conns[i].draw(display);
	}
}
//#end class Point2(Vector2)

/*
 * Class Conn
 * Only exists in JavaScript because of the design of JSON.
 * 
 */
function Conn(a, b)
{
	this.a = a;
	this.b = b;

	// Line which we draw.
	this.visual = null;
	this.visualc = {'visual':this.visual};
}
// Turn this into a class.
new Conn();

Conn.prototype.resolve = function(map)
{
	if(typeof this.a == "string")
	{
		this.a = map.getNamed(this.a);
		this.b = map.getNamed(this.b);
	}
}

/*
 * Conn::draw() -- update the screen.
 * 
 */
Conn.prototype.draw = function(display)
{
	// Draw it.
	if(this.visual === null)
	{
		// Don't add a line if our complement has already done it.
		if(this.visualc.visual !== null)
		{
			this.visual = this.visualc.visual;
		}
		else
		{
			var svg = 'http://www.w3.org/2000/svg';
			this.visual = document.createElementNS(svg, 'line');
			this.visual.setAttributeNS(null, 'stroke', '#000');
			this.visual.setAttributeNS(null, 'stroke-width', '1');
			this.visual.setAttributeNS(null, 'stroke-opacity', '1');
			this.visual.onmousedown = function() { return false; }
			display.appendChild(this.visual);
			// Connect the complementary's connection to this.
			this.visualc.visual = this.visual;
		}
	}
	// TODO: This shouldn't fire when this.visual == this.visualc.visual...
	this.visual.setAttributeNS(null, 'x1', this.a.x);
	this.visual.setAttributeNS(null, 'y1', this.a.y);
	this.visual.setAttributeNS(null, 'x2', this.b.x);
	this.visual.setAttributeNS(null, 'y2', this.b.y);
}
//#end class Conn

function organize(map)
{
	var f = 0.71;
	var scale = (map.width + map.height) / 2.0;
	//map.randomize(Math.floor(scale / (0.65*f)));
	for(i = 0; i < 12; i++) {
		map.sort2(scale / (7.5*f), 0.7);
		if(i % 30 == 0) {
			map.sort3();
		} //#end if
		if(i % 32 == 0) {
			map.sort4();
		} //#end if
		if (i % 25 == 0) {
			map.randomize(Math.floor(scale / (10.0*f)));
		} //#end if
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
	/*
	*/
	} //#next i
	for(i = 0; i < 12; i++) {
		map.sort4();
	} //#next i
	//#m.sort3()
	map.center();
}

function createMap(seed_id) {
	if(seed_id == undefined) { seed_id = 3243; }
	// FIXME: No random seed in JavaScript, have to implement your own RNG. 
	// No, really.
	//Math.seed(seed_id);
	var a = new Map();
	var b = [];
	var i;
	for(i = 0; i < 8; i++) {
		b.push(new Point(String.fromCharCode(String.charCodeAt('b') + i)))
	} //#next i
	for(i = 0; i < 7; i++) {
		b[i].addConn(b[i+1]);
	} //#next i
	var addConn1 = randint(0,b.length-1);
	var addConn2 = (addConn1 + randint(0,b.length)) % b.length;
	if (addConn1 == addConn2) {
		//# Possible
		addConn2 = (addConn1 ^ 7) % b.length;
	}
	if (addConn1 == addConn2) {
		//# Unlikely
		addConn1 = 0;
	}
	if (addConn1 == addConn2) {
		//# Impossible
		addConn2 = 1;
	}
	//#print b, addConn1, addConn2
	b[addConn1].addConn(b[addConn2]);
	a.addPoints(b);
	for(i = 0; i < 4; i++) {
		a.randomize();
		a.sort1();
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
		//#a.sort1()
	} //#next i
	return a;
} //#end def createMap([seed_id])
