#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SWW Minimize 1
by Javantea
Sept 30, 2015
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

TODO: Better AcceptPotential
"""
from __future__ import print_function
import numpy as np
from scipy.optimize import minimize, basinhopping
import mapmake1
import math
Vector2 = mapmake1.Vector2

def nameIncremental(i):
	alphabet = 'abcdefghijklmnopqrstuvwxyz'
	output = ''
	while i >= 26:
		i, p = divmod(i, 26)
		output += alphabet[p]
	#loop
	output += alphabet[i]
	return output
#end def nameIncremental(i)

def createMap2(nodes=8, conns=7, seed_id=3243):
	connsAvailable = sum(range(nodes)) - (nodes - 1)
	if conns > connsAvailable: raise ValueError("Too many connections, conns >= sum(range(nodes)) - (nodes - 1)")
	mapmake1.seed(seed_id)
	a = mapmake1.Map()
	b = []
	for i in range(nodes):
		b.append(mapmake1.Point(nameIncremental(i)))
	#next i
	# Connect them all end to end
	for i in range(nodes-1):
		b[i].addConn(b[i+1])
	#next i
	for i in range(conns):
		addConn1 = mapmake1.randint(0,len(b)-1)
		# Fully connected!
		while len(b[addConn1].conns) >= (nodes-1):
			addConn1 = mapmake1.randint(0,len(b)-1)
		#end if
		addConn2 = addConn1
		while addConn2 == addConn1 or b[addConn2] in b[addConn1].conns:
			addConn2 = mapmake1.randint(0,len(b)-1)
		#loop
		#print b, addConn1, addConn2
		b[addConn1].addConn(b[addConn2])
	#next i
	a.addPoints(b)
	# Sort it in the way of the branch.
	a.sortBranch()
	# Only 2 pixels of randomization
	a.randomize(2)
	a.center()
	return a
#end def createMap2([nodes], [conns], [seed_id])

class LJ_Potential:
	"""
	Proper Lennard-Jones 6 12 Potential.
	"""
	def __init__(self, r0=40, scale_factor=0.055):
		self.r0 = r0
		self.scale_factor = scale_factor
		self.A = 2 * ((r0)**6) * scale_factor
		self.B = ((r0)**12) * scale_factor
	#end def __init__()
	
	def V_LJ(self, r):
		"""
		Attempt at a Lennard-Jones 6 12 potential
		The potential is very high when the two nodes are close.
		The potential is low when two nodes are far away.
		The potential is lowest when two nodes are at the r0 distance, not too 
		close, not too far away.
		Non-Connected nodes should have a LJ potential.
		r0 is the value for which this LJ potential is optimized.
		"""
		# This is a divide by zero protection.
		if r < 1e-3:
			return self.V_LJ(1e-3)
		#end if
		r6 = 1/(r ** 6)
		return ((-self.A*r6) + (self.B*r6*r6))
	#end def V_LJ(r)
#end class LJ_Potential

def V_Harmonic(r, b0):
	"""
	Attempt at a Harmonic potential
	Connected nodes should have a harmonic potential.
	b0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""
	A = 1
	return A * (r - b0)**2
#end def V_Harmonic(r, b0)

def V_Harmonic_Angle(theta, theta0):
	"""
	Attempt at a Harmonic potential for the angle between two connected items
	Connected nodes should have a harmonic potential.
	theta0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""
	A = 1
	return A * (theta - theta0)**2
#end def V_Harmonic_Angle(theta, theta0)

def V_SmallWideWorld(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential = 0.0
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			Potential += V_Harmonic(dp_len, b0)
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			for i in range(len(conns[i])):
				b = (i+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[i]
				xb = Vector2(0, 0) - dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[i]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					intermediate = top / bot
					# A very stupid floating point bug.
					if intermediate < -1.0:
						theta = math.pi
					elif intermediate > 1.0:
						theta = 0.0
					else:
						theta = math.acos(intermediate)
					#end if
				#end if
				Potential += V_Harmonic_Angle(theta, theta0)
			#next i
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			Potential += lj_conf.V_LJ(dp_len)
		#next j
	#next i
	return Potential
#end def V_SmallWideWorld(x, conns, b0, lj_conf)

def V_SmallWideWorld_verbose(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential = 0.0
	HP = 0.0
	HPA = 0.0
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = V_Harmonic(dp_len, b0)
			print("r: %3.3f, V: %3.3f" % (dp_len, v))
			Potential += v
			HP += v
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			for i in range(len(conns[i])):
				b = (i+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[i]
				xb = Vector2(0, 0) - dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[i]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					#print(top, bot)
					theta = math.acos(top/bot)
				#end if
				v = V_Harmonic_Angle(theta, theta0)
				print("theta: %3.3f, V: %3.3f" % (theta, v))
				Potential += v
				HPA += v
			#next i
		#end if
	#next i
	print("Harmonic Potential: %3.3f" % (HP))
	print("Angular Harmonic Potential: %3.3f" % (HPA))
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = lj_conf.V_LJ(dp_len)
			print("r: %3.3f, V: %3.3f" % (dp_len, v))
			Potential += v
		#next j
	#next i
	print("LJ Potential: %3.3f" % (Potential-(HP+HPA)))
	return Potential
#end def V_SmallWideWorld(x, conns, b0, lj_conf)

def getX0Conns(map1):
	# Get the xy values into a 2D array.
	x0 = np.array([np.array((p.x, p.y)) for p in map1.points])
	# Turn it into a 1D array for the optimizer.
	x0 = x0.reshape(1, 2 * len(map1.points))[0]
	conns = []
	for p in map1.points:
		pConns = []
		for conn in p.conns:
			for i, q in enumerate(map1.points):
				if q == conn:
					pConns.append(i)
					break
				#end if
			#next i, q
		#next conn
		conns.append(pConns)
	#next p
	return x0, conns
#end def getX0Conns(map1)

def getNonConns(conns):
	"""
	Inverse of conns. If a is not connected to b, b should be in conns[a].
	"""
	nonConns = []
	for i in range(len(conns)):
		nc = []
		for j in range(len(conns)):
			# Don't check against ourself.
			if i == j: continue
			# If we are connected, not not connected
			if j in conns[i]: continue
			nc.append(j)
		#next j
		nonConns.append(nc)
	#next i
	return nonConns
#end def getNonConns(conns)
	
def OptimizeBasinHopping(map1, lj_conf, b0 = 40, steps=60, stepsize=50, disp=True, callback=None):
	"""
	Optimize using SciPy BasinHopping and L-BFGS-B
	One step is fairly fast and usually helps, so why not try a step?
	Optimizes in place.
	"""
	x0, conns = getX0Conns(map1)
	nonConns = getNonConns(conns)
	#res = minimize(V_molecular_mult, molecule.x0, args=(molecule.conns,), method='L-BFGS-B',
	#	options={'disp': True}) #, 'T0': 0.1, 'schedule':verbose_fast_sa
	minimizer_kwargs = {'args':(conns, b0, lj_conf, nonConns), 'method':'L-BFGS-B'}
	res = basinhopping(V_SmallWideWorld, x0, minimizer_kwargs=minimizer_kwargs, 
			disp=disp, niter=steps, stepsize=stepsize, callback=callback)

	m = np.matrixlib.matrix(res.x)
	m = m.reshape(len(map1.points), 2)
	for i, xa in enumerate(m):
		x = np.array(xa)[0]
		#print('debug2', x)
		map1.points[i].x = x[0]
		map1.points[i].y = x[1]
	#next i, x
	map1.centerBound()
	return
#end def OptimizeBasinHopping(map1, lj_conf, steps=60, stepsize=50, disp=True, callback=None)

def drawLJ():
	"""
	Draw A Lennard-Jones 6-12 Potential Graph
	"""
	from matplotlib import pyplot
	lj_conf = LJ_Potential()
	data = [lj_conf.V_LJ(x) for x in range(33, 70)]
	xdata = [x for x in range(33, 70)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawLJ()

def drawHarmonic():
	"""
	Draw A Harmonic Potential Graph
	"""
	from matplotlib import pyplot
	b0 = 40
	data = [V_Harmonic(x, b0) for x in range(0, 80)]
	xdata = [x for x in range(0, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawHarmonic()

class AcceptPotential(object):
	def __init__(self, V):
		self.V = V
	def __call__(self, x=None, f=None, accept=None):
		return (f <= self.V)
#end class AcceptPotential(object)

def usage():
	print(main.__doc__)
	return 1
#end def usage()

def main():
	"""Small Wide World Optimize 1
This program creates or optimizes an already created graph using basin-hopping global optimization.
Usage:
sww_optimize1.py [-f filename] [-lin] [nodes] [conns] [steps]
-f filename | Load a JSON file.
-lin        | Take all nodes and line them up from first to last 
              horizontally instead of sortBranch + random or how they were 
              loaded.
nodes       | If you aren't loading from file, how many nodes on the 
              graph. Default: 20
conns       | How many random connections to make between nodes on the 
              backbone. Default: 4
steps       | How many steps to take in the basin-hopping algorithm, 
              determines the quality. Default: 20
"""
	lj_conf = LJ_Potential()
	"""
	#This is how to test the LJ and the Harmonic potentials.
	#aconf = LJConf(A, B)
	#data = [E_Harmonic(x*0.1, a) for x in range(7, 20)]
	data = [V_LJ(x*1.0) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data)
		pyplot.show()
	#end if
	data2 = [V_Harmonic(x*1.0, 40) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data2)
		pyplot.show()
	#end if
	"""
	import sys
	import os
	
	nodes = 20
	conns = 4
	steps = 20
	filename = None
	lin = False
	# When you finish a round and it failed to find the lowest potential, running again is as easy as:
	# mkdir redo && cd redo
	# for i in 10 13 15 16; do python3 ../../sww_optimize1.py -f ../sww_optimize1out2"$i".json; done
	if '-h' in sys.argv:
		return usage()
	#end if
	if '-f' in sys.argv:
		fpos = sys.argv.index('-f')+1
		if fpos >= len(sys.argv): return usage()
		filename = sys.argv[fpos]
		sys.argv.pop(fpos)
		sys.argv.remove('-f')
	#end if
	if '-lin' in sys.argv:
		lin = True
		sys.argv.remove('-lin')
	#end if
	if len(sys.argv) > 1:
		nodes = int(sys.argv[1])
		if len(sys.argv) > 2:
			conns = int(sys.argv[2])
			if len(sys.argv) > 3:
				steps = int(sys.argv[3])
			#end if
		#end if
	#end if
	
	if filename:
		mapmake1.seed(9999)
		map1 = mapmake1.Map()
		map1.loadJson(open(filename, 'r').read())
		nodes = len(map1.points)
		# If we need conns at some point:
		# All connections minus the backbone connections (nodes - 1)
		#conns = sum([len(n.conns) for n in map1.points]) - (nodes - 1)
		map1.desc = "SWW Optimize map with {} nodes from {} [results]".format(nodes, filename)
	else:
		map1 = createMap2(nodes, conns, 9999)
		map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
	#end if
	dri = map1.points[0] - map1.points[1]
	x0, conns = getX0Conns(map1)
	if lin:
		# Perfectly horizontal spaces 40 apart.
		x0a = []
		j = 0
		curr_x = 10
		for x in x0:
			if j == 0:
				x0a.append(curr_x)
				curr_x += 40
				j = 1
			else:
				x0a.append(10)
				j = 0
			#end if
		#next x
		x0 = x0a
		m = np.matrixlib.matrix(x0)
		m = m.reshape(len(x0)//2, 2)
		for i, xa in enumerate(m):
			x = np.array(xa)[0]
			#print('debug2', x)
			map1.points[i].x = x[0]
			map1.points[i].y = x[1]
		#next i, x
	#end if
	nonConns = getNonConns(conns)

	# FIXME: Race condition is very likely due to multiple programs running at once.
	file_num = 0
	while os.path.exists('sww_optimize1out2%ia.svg' % file_num):
		file_num += 1
	#loop
	open('sww_optimize1out2%ia.svg' % file_num, 'w').write(str(map1))
	# See r0_40_nodes_potential.dat
	callback = AcceptPotential(0.15 * nodes * nodes)
	OptimizeBasinHopping(map1, lj_conf, 40, steps, 50, True, callback)
	
	print('sww_optimize1out2%i.svg' % file_num)
	print('sww_optimize1out2%i.json' % file_num)
	print("Configuration:")
	x_end = getX0Conns(map1)[0]
	print(x_end)
	Vmin = V_SmallWideWorld(x_end, conns, 40, lj_conf, nonConns)
	print("Minimum potential:", Vmin)
	dr = Vector2(*(x_end[:2])) - Vector2(*(x_end[2:4]))
	print("Distance:", dr.length())
	#x0 = []
	#for v in map1.points:
	#	x0.extend(v)
	#next v
	#x0 = np.array(x0)
	V0 = V_SmallWideWorld(x0, conns, 40, lj_conf, nonConns)
	print("Initial Potential:", V0)
	print("Initial Distance:", dri.length())
	#print('debug', np.array(m[0])[0], np.array(m[1])[0])
	s = str(map1)
	results = '\nConfiguration:\n{}\nMinimum potential:{}\nDistance: {}\nInitial potential:{}\nInitial Distance: {}\n'.format(x_end, Vmin, dr.length(), V0, dri.length())
	s = s.replace('[results]', results)
	open('sww_optimize1out2%i.svg' % file_num, 'w').write(s)
	open('sww_optimize1out2%i.json' % file_num, 'w').write(map1.saveJson())
	#print("Minimum potential breakdown:")
	#V_SmallWideWorld_verbose(res.x, conns, 40, lj_conf)
#end def main()

if __name__ == '__main__':
    main()
#end if
