#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SWW Minimize 1
by Javantea
Sept 30, 2015
Based on N Atom molecule using Basin-Hopping Method (aka L-BFGS-B)
by Javantea
July 23, 2014
Based on metropolis_min1.py by Javantea, July 17-22, 2014
Based on s_4atom.py by Javantea, June 24, 2014

# TODO: Improve Lennard-Jones for non-connected nodes.
"""
from __future__ import print_function
try:
	from matplotlib import pyplot
except ImportError as ie:
	print("Matplotlib is an optional dependency if you want graphical validation of LJ potential.")
	pyplot = None
#end try
import numpy as np
from scipy.optimize import minimize, basinhopping
import mapmake1
import math
Vector2 = mapmake1.Vector2

def nameIncremental(i):
	alphabet = 'abcdefghijklmnopqrstuvwxyz'
	output = ''
	while i >= 26:
		i, p = divmod(i, 26)
		output += alphabet[p]
	#loop
	output += alphabet[i]
	return output
#end def nameIncremental(i)

def createMap2(nodes=8, conns=7, seed_id=3243):
	connsAvailable = sum(range(nodes)) - (nodes - 1)
	if conns > connsAvailable: raise ValueError("Too many connections, conns >= sum(range(nodes)) - (nodes - 1)")
	mapmake1.seed(seed_id)
	a = mapmake1.Map()
	b = []
	for i in range(nodes):
		b.append(mapmake1.Point(nameIncremental(i)))
	#next i
	# Connect them all end to end
	for i in range(nodes-1):
		b[i].addConn(b[i+1])
	#next i
	for i in range(conns):
		addConn1 = mapmake1.randint(0,len(b)-1)
		# Fully connected!
		while len(b[addConn1].conns) >= (nodes-1):
			addConn1 = mapmake1.randint(0,len(b)-1)
		#end if
		addConn2 = addConn1
		while addConn2 == addConn1 or b[addConn2] in b[addConn1].conns:
			addConn2 = mapmake1.randint(0,len(b)-1)
		#loop
		#print b, addConn1, addConn2
		b[addConn1].addConn(b[addConn2])
	#next i
	a.addPoints(b)
	a.randomize()
	a.center()
	return a
#end def createMap2([nodes], [conns], [seed_id])

class LJ_Potential:
	"""
	Proper Lennard-Jones 6 12 Potential.
	"""
	def __init__(self, r0=140, scale_factor=0.055):
		self.r0 = r0
		self.scale_factor = scale_factor
		self.A = 2 * ((r0)**6) * scale_factor
		self.B = ((r0)**12) * scale_factor
	#end def __init__()
	
	def V_LJ(self, r):
		"""
		Attempt at a Lennard-Jones 6 12 potential
		The potential is very high when the two nodes are close.
		The potential is low when two nodes are far away.
		The potential is lowest when two nodes are at the r0 distance, not too 
		close, not too far away.
		Non-Connected nodes should have a LJ potential.
		r0 is the value for which this LJ potential is optimized.
		"""
		# This is a divide by zero protection.
		if r < 1e-3:
			return self.V_LJ(1e-3)
		#end if
		r6 = 1/(r ** 6)
		return ((-self.A*r6) + (self.B*r6*r6))
	#end def V_LJ(r)
#end class LJ_Potential

def V_Harmonic(r, b0):
	"""
	Attempt at a Harmonic potential
	Connected nodes should have a harmonic potential.
	b0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""
	A = 1
	return A * (r - b0)**2
#end def V_Harmonic(r, b0)

def V_Harmonic_Angle(theta, theta0):
	"""
	Attempt at a Harmonic potential for the angle between two connected items
	Connected nodes should have a harmonic potential.
	theta0 is the value for which this Harmonic potential is optimized.
	A is the scale of potential.
	"""
	A = 1
	return A * (theta - theta0)**2
#end def V_Harmonic_Angle(theta, theta0)

def V_SmallWideWorld(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential = 0.0
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			Potential += V_Harmonic(dp_len, b0)
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			for i in range(len(conns[i])):
				b = (i+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[i]
				xb = Vector2(0, 0) - dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[i]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					intermediate = top / bot
					# A very stupid floating point bug.
					if intermediate < -1.0:
						theta = math.pi
					elif intermediate > 1.0:
						theta = 0.0
					else:
						theta = math.acos(intermediate)
					#end if
				#end if
				Potential += V_Harmonic_Angle(theta, theta0)
			#next i
		#end if
	#next i
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = lj_conf.V_LJ(dp_len)
			# only positive potentials.
			if v > 0:
				Potential += v
			#end if
		#next j
	#next i
	return Potential
#end def V_SmallWideWorld(x, conns, b0, lj_conf)

def V_SmallWideWorld_verbose(x, conns, b0, lj_conf, nonConns):
	"""
	Just like a molecule but in 2D.
	"""
	Potential = 0.0
	HP = 0.0
	HPA = 0.0
	# Harmonic potential with length b0.
	for i in range(len(conns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		verts = [pt_i]
		dps = []
		dp_lens = []
		for j in conns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = V_Harmonic(dp_len, b0)
			print("r: %3.3f, V: %3.3f" % (dp_len, v))
			Potential += v
			HP += v
			verts.append(pt_j)
			dps.append(dp)
			dp_lens.append(dp_len)
		#next j
		if len(verts) > 2:
			# Angular Harmonic calculation
			theta0 = 2 * math.pi / len(conns[i])
			# theta is computed for axb, bxc, cxa
			for i in range(len(conns[i])):
				b = (i+1) % len(dps)
				# Theta between axb is arccos((ax·xb)/(|ax|*|xb|))
				ax = dps[i]
				xb = Vector2(0, 0) - dps[b]
				#print('xb', '(%3.3f,%3.3f)' % (xb.x, xb.y), '(%3.3f,%3.3f)' % (dps[b].x, dps[b].y), xb.length(), dp_lens[b], dps[b].length())
				top = ax.dot(xb)
				bot = dp_lens[i]*dp_lens[b]
				if bot == 0:
					theta = 0
				else:
					#print(top, bot)
					theta = math.acos(top/bot)
				#end if
				v = V_Harmonic_Angle(theta, theta0)
				print("theta: %3.3f, V: %3.3f" % (theta, v))
				Potential += v
				HPA += v
			#next i
		#end if
	#next i
	print("Harmonic Potential: %3.3f" % (HP))
	print("Angular Harmonic Potential: %3.3f" % (HPA))
	# Lennard-Jones for non-connected nodes.
	for i in range(len(nonConns)):
		pt_i = Vector2(*x[2*i:2*i+2])
		for j in nonConns[i]:
			pt_j = Vector2(*x[2*j:2*j+2])
			dp = pt_i - pt_j
			dp_len = dp.length()
			v = lj_conf.V_LJ(dp_len)
			print("r: %3.3f, V: %3.3f" % (dp_len, v))
			if v > 0:
				Potential += v
			#end if
		#next j
	#next i
	print("LJ Potential: %3.3f" % (Potential-(HP+HPA)))
	return Potential
#end def V_SmallWideWorld(x, conns, b0, lj_conf)

def getX0Conns(map1):
	# Get the xy values into a 2D array.
	x0 = np.array([np.array((p.x, p.y)) for p in map1.points])
	# Turn it into a 1D array for the optimizer.
	x0 = x0.reshape(1, 2 * len(map1.points))[0]
	conns = []
	for p in map1.points:
		pConns = []
		for conn in p.conns:
			for i, q in enumerate(map1.points):
				if q == conn:
					pConns.append(i)
					break
				#end if
			#next i, q
		#next conn
		conns.append(pConns)
	#next p
	return x0, conns
#end def getX0Conns(map1)

def getNonConns(conns):
	"""
	Inverse of conns. If a is not connected to b, b should be in conns[a].
	"""
	nonConns = []
	for i in range(len(conns)):
		nc = []
		for j in range(len(conns)):
			# Don't check against ourself.
			if i == j: continue
			# If we are connected, not not connected
			if j in conns[i]: continue
			nc.append(j)
		#next j
		nonConns.append(nc)
	#next i
	return nonConns
#end def getNonConns(conns)
	
def OptimizeBasinHopping(map1, lj_conf, b0 = 40, steps=60, stepsize=50, disp=True, callback=None):
	"""
	Optimize using SciPy BasinHopping and L-BFGS-B
	One step is fairly fast and usually helps, so why not try a step?
	Optimizes in place.
	"""
	x0, conns = getX0Conns(map1)
	nonConns = getNonConns(conns)
	#res = minimize(V_molecular_mult, molecule.x0, args=(molecule.conns,), method='L-BFGS-B',
	#	options={'disp': True}) #, 'T0': 0.1, 'schedule':verbose_fast_sa
	minimizer_kwargs = {'args':(conns, b0, lj_conf, nonConns), 'method':'L-BFGS-B'}
	res = basinhopping(V_SmallWideWorld, x0, minimizer_kwargs=minimizer_kwargs, 
			disp=disp, niter=steps, stepsize=stepsize, callback=callback)

	m = np.matrixlib.matrix(res.x)
	m = m.reshape(len(map1.points), 2)
	for i, xa in enumerate(m):
		x = np.array(xa)[0]
		#print('debug2', x)
		map1.points[i].x = x[0]
		map1.points[i].y = x[1]
	#next i, x
	map1.centerBound()
	return
#end def OptimizeBasinHopping(map1, lj_conf, steps=60, stepsize=50, disp=True, callback=None)

def drawLJ():
	from matplotlib import pyplot
	lj_conf = LJ_Potential()
	data = [lj_conf.V_LJ(x) for x in range(33, 70)]
	xdata = [x for x in range(33, 70)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawLJ()

def drawHarmonic():
	from matplotlib import pyplot
	b0 = 40
	data = [V_Harmonic(x, b0) for x in range(0, 80)]
	xdata = [x for x in range(0, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		fig = pyplot.figure()
		ax = fig.add_subplot(111)
		ax.set_xlabel('distance (pixels)')
		ax.set_ylabel('Potential (arbitrary magnitude)')
		ax.plot(xdata, data)
		#ax.axis([0.7, 2, 0, None])
		pyplot.show()
	#end if
#end def drawHarmonic()

class AcceptPotential(object):
	def __init__(self, V):
		self.V = V
	def __call__(self, x=None, f=None, accept=None):
		return (f <= self.V)
#end class AcceptPotential(object)

def main():
	lj_conf = LJ_Potential()
	"""
	#This is how to test the LJ and the Harmonic potentials.
	#aconf = LJConf(A, B)
	#data = [E_Harmonic(x*0.1, a) for x in range(7, 20)]
	data = [V_LJ(x*1.0) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data)
		pyplot.show()
	#end if
	data2 = [V_Harmonic(x*1.0, 40) for x in range(30, 80)]
	#print('\n'.join(['%3.3f %3.3f' % ((x*0.01) + 1, E) for (x, E) in enumerate(data)]))
	
	if pyplot:
		# TODO: x axis.
		pyplot.plot(data2)
		pyplot.show()
	#end if
	"""
	import sys
	import os
	
	nodes = 20
	conns = 4
	steps = 40
	x1 = False
	lin = False
	if '-x1' in sys.argv:
		x1 = True
		sys.argv.remove('-x1')
	#end if
	if '-lin' in sys.argv:
		lin = True
		sys.argv.remove('-lin')
	#end if
	if len(sys.argv) > 1:
		nodes = int(sys.argv[1])
		if len(sys.argv) > 2:
			conns = int(sys.argv[2])
			if len(sys.argv) > 3:
				steps = int(sys.argv[3])
			#end if
		#end if
	#end if
	
	map1 = createMap2(nodes, conns, 9999)
	map1.desc = "SWW Optimize map with {} nodes and {} conns [results]".format(nodes, conns)
	dri = map1.points[0] - map1.points[1]
	x0, conns = getX0Conns(map1)
	if x1 or lin:
		if x1:
			#x0 = [510.81305301, 385.65147798, 469.75645727, 392.61197485, 427.16541225, 396.01954412, 382.81828934, 393.89912569, 345.00536839, 411.11936337, 346.75667239, 369.61277072, 328.12245099, 329.2921495,  315.22425258, 288.763412, 306.39832123, 246.89207769, 293.30697898, 206.19322215, 276.1015715,  167.04972025, 254.96226834, 129.8842716,  237.75194095,  90.73870705, 224.65026922,  50.06737728, 207.25813946,  10.80185702, 193.9085541,  -28.59705128]
			x0 = [ 147.56420427,  43.79795302, 189.19732683,  44.68727509, 231.66791435,  49.36023504, 274.82081652,  59.80144291, 315.20266694,  50.02125504, 305.65524673,  90.4528723,  316.34993008, 133.56425343, 321.36519958, 175.79901571, 330.62383549, 217.57657446, 344.10700811, 258.14674393, 361.66757177, 297.13241723, 375.10795718, 337.72261268, 384.29768979, 379.48027738, 389.142904, 421.96628266, 389.6194682,  464.6928152, 385.60133738, 507.44900344, 378.00892276, 548.3489705 ]
		else:
			# Perfectly horizontal spaces 40 apart.
			x0a = []
			j = 0
			curr_x = 10
			for x in x0:
				if j == 0:
					x0a.append(curr_x)
					curr_x += 40
					j = 1
				else:
					x0a.append(10)
					j = 0
				#end if
			#next x
			x0 = x0a
		#end if
		m = np.matrixlib.matrix(x0)
		m = m.reshape(len(x0)//2, 2)
		for i, xa in enumerate(m):
			x = np.array(xa)[0]
			#print('debug2', x)
			map1.points[i].x = x[0]
			map1.points[i].y = x[1]
		#next i, x
	#end if
	nonConns = getNonConns(conns)
	#print('Conns:', conns)

	# FIXME: Race condition is very likely due to multiple programs running at once.
	file_num = 0
	while os.path.exists('sww_optimize1out2%ia.svg' % file_num):
		file_num += 1
	#loop
	open('sww_optimize1out2%ia.svg' % file_num, 'w').write(str(map1))
	callback = AcceptPotential((81.956*nodes) -178.76)
	OptimizeBasinHopping(map1, lj_conf, 40, steps, 50, True, callback)
	
	print('sww_optimize1out2%i.svg' % file_num)
	print('sww_optimize1out2%i.json' % file_num)
	print("Configuration:")
	x_end = getX0Conns(map1)[0]
	print(x_end)
	Vmin = V_SmallWideWorld(x_end, conns, 40, lj_conf, nonConns)
	print("Minimum potential:", Vmin)
	dr = Vector2(*(x_end[:2])) - Vector2(*(x_end[2:4]))
	print("Distance:", dr.length())
	#x0 = []
	#for v in map1.points:
	#	x0.extend(v)
	#next v
	#x0 = np.array(x0)
	V0 = V_SmallWideWorld(x0, conns, 40, lj_conf, nonConns)
	print("Initial Potential:", V0)
	print("Initial Distance:", dri.length())
	#print('debug', np.array(m[0])[0], np.array(m[1])[0])
	s = str(map1)
	results = '\nConfiguration:\n{}\nMinimum potential:{}\nDistance: {}\nInitial potential:{}\nInitial Distance: {}\n'.format(x_end, Vmin, dr.length(), V0, dri.length())
	s = s.replace('[results]', results)
	open('sww_optimize1out2%i.svg' % file_num, 'w').write(s)
	open('sww_optimize1out2%i.json' % file_num, 'w').write(map1.saveJson())
	#print("Minimum potential breakdown:")
	#V_SmallWideWorld_verbose(res.x, conns, 40, lj_conf)
#end def main()

if __name__ == '__main__':
    main()
#end if
